#define _GNU_SOURCE
#include <math.h>
#include <err.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <sys/time.h>
#include <time.h>
#include <limits.h>
#include "stdrusty.h"
#include "list/list.h"
#include <SDL/SDL.h>
#include <cwiid.h>

#define MAX_X CWIID_IR_X_MAX
#define MAX_Y CWIID_IR_Y_MAX

/* Obtained from http://tog.acm.org/GraphicsGems/gemsiii/insectc.c */
/* Faster Line Segment Intersection   */
/* Franklin Antonio                   */

/* The use of some short working variables allows this code to run   */
/* faster on 16-bit computers, but is not essential.  It should not  */
/* affect operation on 32-bit computers.  The short working variables*/
/* to not restrict the range of valid input values, as these were    */
/* constrained in any case, due to algorithm restrictions.           */
static bool lines_intersect(double x1, double y1,
			    double x2, double y2,
			    double x3, double y3,
			    double x4, double y4,
			    double *x, double *y) 
{

	double Ax,Bx,Cx,Ay,By,Cy,d,e,f,num;
	double x1lo,x1hi,y1lo,y1hi;

	Ax = x2-x1;
	Bx = x3-x4;

	if(Ax<0) {						/* X bound box test*/
		x1lo=x2; x1hi=x1;
	} else {
		x1hi=x2; x1lo=x1;
	}
	if(Bx>0) {
		if(x1hi < x4 || x3 < x1lo) return false;
	} else {
		if(x1hi < x3 || x4 < x1lo) return false;
	}

	Ay = y2-y1;
	By = y3-y4;

	if(Ay<0) {						/* Y bound box test*/
		y1lo=y2; y1hi=y1;
	} else {
		y1hi=y2; y1lo=y1;
	}
	if(By>0) {
		if(y1hi < y4 || y3 < y1lo) return false;
	} else {
		if(y1hi < y3 || y4 < y1lo) return false;
	}


	Cx = x1-x3;
	Cy = y1-y3;
	d = By*Cx - Bx*Cy;					/* alpha numerator*/
	f = Ay*Bx - Ax*By;					/* both denominator*/
	if(f>0) {						/* alpha tests*/
		if(d<0 || d>f) return false;
	} else {
		if(d>0 || d<f) return false;
	}

	e = Ax*Cy - Ay*Cx;					/* beta numerator*/
	if(f>0) {						/* beta tests*/
		if(e<0 || e>f) return false;
	} else {
		if(e>0 || e<f) return false;
	}

/*compute intersection coordinates*/

	if(f==0) return false;
	num = d*Ax;						/* numerator */
	*x = x1 + num / f;				/* intersection x */

	num = d*Ay;
	*y = y1 + num / f;				/* intersection y */

	return true;
}

// A set of very useful macros that you will find in most
// code that I write whether I use them in a program or
// not.

#define abs(a) (((a)<0) ? -(a) : (a))
#define sign(a) (((a)<0) ? -1 : (a)>0 ? 1 : 0)

// The following code implements a Bresenham line drawing
// algorithm. There are 4 separate routines each optimized
// for one of the four pixel depths supported by SDL. SDL
// support many pixel formats, but it only support 8, 16,
// 24, and 32 bit pixels.

//----------------------------------------------------------

// Draw lines in 8 bit surfaces.

static void line8(SDL_Surface *s, 
                  int x1, int y1, 
                  int x2, int y2, 
                  Uint32 color)
{
  int d;
  int x;
  int y;
  int ax;
  int ay;
  int sx;
  int sy;
  int dx;
  int dy;

  Uint8 *lineAddr;
  Sint32 yOffset;

  dx = x2 - x1;  
  ax = abs(dx) << 1;  
  sx = sign(dx);

  dy = y2 - y1;  
  ay = abs(dy) << 1;  
  sy = sign(dy);
  yOffset = sy * s->pitch;

  x = x1;
  y = y1;

  lineAddr = ((Uint8 *)(s->pixels)) + (y * s->pitch);
  if (ax>ay)
  {                      /* x dominant */
    d = ay - (ax >> 1);
    for (;;)
    {
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        *(lineAddr + x) = (Uint8)color;

      if (x == x2)
      {
        return;
      }
      if (d>=0)
      {
        y += sy;
        lineAddr += yOffset;
        d -= ax;
      }
      x += sx;
      d += ay;
    }
  }
  else
  {                      /* y dominant */
    d = ax - (ay >> 1);
    for (;;)
    {
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        *(lineAddr + x) = (Uint8)color;

      if (y == y2)
      {
        return;
      }
      if (d>=0) 
      {
        x += sx;
        d -= ay;
      }
      y += sy;
      lineAddr += yOffset;
      d += ax;
    }
  }
}

//----------------------------------------------------------

// Draw lines in 16 bit surfaces. Note that this code will
// also work on 15 bit surfaces.

static void line16(SDL_Surface *s, 
                   int x1, int y1, 
                   int x2, int y2, 
                   Uint32 color)
{
  int d;
  int x;
  int y;
  int ax;
  int ay;
  int sx;
  int sy;
  int dx;
  int dy;

  Uint8 *lineAddr;
  Sint32 yOffset;

  dx = x2 - x1;  
  ax = abs(dx) << 1;  
  sx = sign(dx);

  dy = y2 - y1;  
  ay = abs(dy) << 1;  
  sy = sign(dy);
  yOffset = sy * s->pitch;

  x = x1;
  y = y1;

  lineAddr = ((Uint8 *)s->pixels) + (y * s->pitch);
  if (ax>ay)
  {                      /* x dominant */
    d = ay - (ax >> 1);
    for (;;)
    {
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        *((Uint16 *)(lineAddr + (x << 1))) = (Uint16)color;

      if (x == x2)
      {
        return;
      }
      if (d>=0)
      {
        y += sy;
        lineAddr += yOffset;
        d -= ax;
      }
      x += sx;
      d += ay;
    }
  }
  else
  {                      /* y dominant */
    d = ax - (ay >> 1);
    for (;;)
    {
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        *((Uint16 *)(lineAddr + (x << 1))) = (Uint16)color;

      if (y == y2)
      {
        return;
      }
      if (d>=0) 
      {
        x += sx;
        d -= ay;
      }
      y += sy;
      lineAddr += yOffset;
      d += ax;
    }
  }
}

//----------------------------------------------------------

// Draw lines in 24 bit surfaces. 24 bit surfaces require
// special handling because the pixels don't fall on even
// address boundaries. Instead of being able to store a
// single byte, word, or long you have to store 3
// individual bytes. As a result 24 bit graphics is slower
// than the other pixel sizes.

static void line24(SDL_Surface *s, 
                   int x1, int y1, 
                   int x2, int y2, 
                   Uint32 color)
{
  int d;
  int x;
  int y;
  int ax;
  int ay;
  int sx;
  int sy;
  int dx;
  int dy;

  Uint8 *lineAddr;
  Sint32 yOffset;

#if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
  color <<= 8;
#endif

  dx = x2 - x1;  
  ax = abs(dx) << 1;  
  sx = sign(dx);

  dy = y2 - y1;  
  ay = abs(dy) << 1;  
  sy = sign(dy);
  yOffset = sy * s->pitch;

  x = x1;
  y = y1;

  lineAddr = ((Uint8 *)(s->pixels)) + (y * s->pitch);
  if (ax>ay)
  {                      /* x dominant */
    d = ay - (ax >> 1);
    for (;;)
    {
      Uint8 *p = (lineAddr + (x * 3));
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        memcpy(p, &color, 3);

      if (x == x2)
      {
        return;
      }
      if (d>=0)
      {
        y += sy;
        lineAddr += yOffset;
        d -= ax;
      }
      x += sx;
      d += ay;
    }
  }
  else
  {                      /* y dominant */
    d = ax - (ay >> 1);
    for (;;)
    {
      Uint8 *p = (lineAddr + (x * 3));
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        memcpy(p, &color, 3);

      if (y == y2)
      {
        return;
      }
      if (d>=0) 
      {
        x += sx;
        d -= ay;
      }
      y += sy;
      lineAddr += yOffset;
      d += ax;
    }
  }
}

//----------------------------------------------------------

// Draw lines in 32 bit surfaces. Note that this routine
// ignores alpha values. It writes them into the surface
// if they are included in the pixel, but does nothing
// else with them.

static void line32(SDL_Surface *s, 
                   int x1, int y1, 
                   int x2, int y2, 
                   Uint32 color)
{
  int d;
  int x;
  int y;
  int ax;
  int ay;
  int sx;
  int sy;
  int dx;
  int dy;

  Uint8 *lineAddr;
  Sint32 yOffset;

  dx = x2 - x1;  
  ax = abs(dx) << 1;  
  sx = sign(dx);

  dy = y2 - y1;  
  ay = abs(dy) << 1;  
  sy = sign(dy);
  yOffset = sy * s->pitch;

  x = x1;
  y = y1;

  lineAddr = ((Uint8 *)(s->pixels)) + (y * s->pitch);
  if (ax>ay)
  {                      /* x dominant */
    d = ay - (ax >> 1);
    for (;;)
    {
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        *((Uint32 *)(lineAddr + (x << 2))) = (Uint32)color;

      if (x == x2)
      {
        return;
      }
      if (d>=0)
      {
        y += sy;
        lineAddr += yOffset;
        d -= ax;
      }
      x += sx;
      d += ay;
    }
  }
  else
  {                      /* y dominant */
    d = ax - (ay >> 1);
    for (;;)
    {
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        *((Uint32 *)(lineAddr + (x << 2))) = (Uint32)color;

      if (y == y2)
      {
        return;
      }
      if (d>=0) 
      {
        x += sx;
        d -= ay;
      }
      y += sy;
      lineAddr += yOffset;
      d += ax;
    }
  }
}

//----------------------------------------------------------

// Examine the depth of a surface and select a line
// drawing routine optimized for the bytes/pixel of the
// surface.

static void line(SDL_Surface *s, 
                 int x1, int y1, 
                 int x2, int y2, 
                 Uint32 color)
{
  switch (s->format->BytesPerPixel)
  {
  case 1:
    line8(s, x1, y1, x2, y2, color);
    break;
  case 2:
    line16(s, x1, y1, x2, y2, color);
    break;
  case 3:
    line24(s, x1, y1, x2, y2, color);
    break;
  case 4:
    line32(s, x1, y1, x2, y2, color);
    break;
  }
}

#if 1
struct coord
{
	double x;
	double y;
};

struct ball
{
	struct coord pos;
	struct coord move;
};

struct line_segment
{
	struct coord start, end;
	struct list_node list;
	bool ignore;
	struct timeval expires;
};

static void thick_line(SDL_Surface *s, const struct line_segment *l,
		       unsigned color)
{
	struct line_segment adj;
	int x, y, minx = MAX_X, miny = MAX_Y, maxx = 0, maxy = 0;

	SDL_LockSurface(s);

	/* Cheap hack */
	for (x = -1; x < 2; x++) {
		for (y = -1; y < 2; y++) {
			adj.start.x = l->start.x + x;
			adj.end.x = l->end.x + x;
			adj.start.y = l->start.y + y;
			adj.end.y = l->end.y + y;

			if (adj.start.x < minx)
				minx = adj.start.x;
			if (adj.start.x > maxx)
				maxx = adj.start.x;
			if (adj.start.y < miny)
				miny = adj.start.y;
			if (adj.start.y > maxy)
				maxy = adj.start.y;
			if (adj.end.x < minx)
				minx = adj.end.x;
			if (adj.end.x > maxx)
				maxx = adj.end.x;
			if (adj.end.y < miny)
				miny = adj.end.y;
			if (adj.end.y > maxy)
				maxy = adj.end.y;

			line(s, adj.start.x, adj.start.y, adj.end.x, adj.end.y,
			     color);
		}
	}

	if (minx < 0)
		minx = 0;
	if (miny < 0)
		miny = 0;
	if (maxx >= MAX_X)
		maxx = MAX_X-1;
	if (maxy >= MAX_Y)
		maxy = MAX_Y-1;
	SDL_UpdateRect(s, minx, miny, maxx-minx, maxy-miny);
	SDL_UnlockSurface(s);
}

static bool intersect(const struct coord *c1, const struct coord *c2,
		      const struct line_segment *seg, struct coord *intersect)
{
	return lines_intersect(c1->x, c1->y, c2->x, c2->y,
			       seg->start.x, seg->start.y,
			       seg->end.x, seg->end.y,
			       &intersect->x, &intersect->y);
}

static double dist(const struct coord *c1, const struct coord *c2)
{
	float x = (c1->x - c2->x), y = (c1->y - c2->y);
	return sqrt(x * x + y * y);
}

static unsigned rad_to_deg(double rad)
{
	return ((unsigned)(rad / M_PI * 180) + 360) % 360;
}

static void bounce(struct ball *ball, const struct line_segment *line,
		   struct coord *remainder)
{
	double len, speed, lang, iang, oang;
	struct coord isect, new;

	new.x = ball->pos.x + ball->move.x;
	new.y = ball->pos.y + ball->move.y;

	/* How far were we supposed to move ball? */
	len = sqrt(remainder->x * remainder->x + remainder->y * remainder->y);

	/* How far is it to intersection? */
	intersect(&ball->pos, &new, line, &isect);
	len -= dist(&ball->pos, &isect);

	/* Move ball to intersection. */
	ball->pos = isect;
	printf("ball is now at %f,%f\n", ball->pos.x, ball->pos.y);

	/* Outgoing angle = 2 * line angle - incident angle. */
	lang = atan2(line->end.x - line->start.x, line->end.y - line->start.y);
	iang = atan2(ball->move.x, ball->move.y);
	oang = 2 * lang - iang, 2*M_PI;
	printf("lang = %u, iang = %u, oang=%u\n", 
	       rad_to_deg(lang),
	       rad_to_deg(iang),
	       rad_to_deg(oang));

	/* Set new direction for ball, at same speed */
	speed = sqrt(ball->move.x * ball->move.x + ball->move.y * ball->move.y);
	speed *= 1.01;
	ball->move.x = sin(oang) * speed;
	ball->move.y = cos(oang) * speed;

	/* Set remainder. */
	remainder->x = sin(oang) * len;
	remainder->y = cos(oang) * len;
	printf("len = %f, remainder = %f,%f\n", len, remainder->x, remainder->y);
}

static struct line_segment border[] = {
	{ { 0, 0, }, { MAX_X-1, 0 } },
	{ { MAX_X-1, 0, }, { MAX_X-1, MAX_Y-1 } },
	{ { MAX_X-1, MAX_Y-1, }, { 0, MAX_Y-1 } },
	{ { 0, MAX_Y-1, }, { 0, 0 } },
};

static inline float deg_to_rad(float degrees)
{
	return degrees * M_PI / 180;
}

static SDL_Surface *sub_surface(SDL_Surface *screen, int w, int h)
{
	return SDL_CreateRGBSurface(SDL_SWSURFACE, w, h,
				    screen->format->BitsPerPixel,
				    screen->format->Rmask,
				    screen->format->Gmask,
				    screen->format->Bmask,
				    screen->format->Amask);
}

static SDL_Surface *ball_surface(SDL_Surface *screen)
{
	SDL_Surface *ball;

	/* Just like the screen surface. */
	ball = sub_surface(screen, 5, 5);

	/* Lock the surface */
	SDL_LockSurface(ball);

	/* Black, white corners. */
	memset(ball->pixels, 0x00, ball->pitch * 5);
	line(ball, 0, 0, 0, 0, 0xFFFFFFFF);
	line(ball, 4, 0, 4, 0, 0xFFFFFFFF);
	line(ball, 0, 4, 0, 4, 0xFFFFFFFF);
	line(ball, 4, 4, 4, 4, 0xFFFFFFFF);
	SDL_UnlockSurface(ball);

	return ball;
}

static void clear_ignore(struct list_head *lines)
{
	struct line_segment *i;
	printf("Unignoring...\n");
	list_for_each(lines, i, list)
		i->ignore = false;
}

int main(int argc, char *argv[])
{
	struct ball ball;
	SDL_Surface *screen, *ballsur, *savesur = NULL;
	Uint8  video_bpp;
	Uint32 videoflags;
	unsigned int i;
	SDL_Rect rect;
	bool creating_lines = false;
	struct timeval line_life = { 5, 0 };
	LIST_HEAD(lines);

	if (argc != 5)
		errx(1, "Usage: %s ballx bally ballangle ballspeed\n"
		     "   Where ballangle is 0 for north, 90 for east, etc\n",
			argv[0]);

	ball.pos.x = atol(argv[1]);
	ball.pos.y = atol(argv[2]);
	ball.move.x = sin(deg_to_rad(atof(argv[3]))) * atol(argv[4]);
	ball.move.y = cos(deg_to_rad(atof(argv[3]))) * atol(argv[4]);
	printf("ball move = %f,%f\n", ball.move.x, ball.move.y);

	/* Initialize SDL */
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Couldn't initialize SDL: %s\n",SDL_GetError());
		return(1);
	}

	video_bpp = 0;
	videoflags = SDL_SWSURFACE;

	/* Set 640x480 video mode */
	if ( (screen=SDL_SetVideoMode(MAX_X,MAX_Y,video_bpp,videoflags)) == NULL ) {
		errx(1, "Couldn't set %dx%dx%d video mode: %s",
		     MAX_X, MAX_Y,video_bpp, SDL_GetError());
	}

	/* Set the surface pixels and refresh! */
	if ( SDL_LockSurface(screen) < 0 ) {
		errx(1, "Couldn't lock the display surface: %s",
		     SDL_GetError());
	}
	/* White. */
	memset(screen->pixels, 0xFF, screen->pitch * screen->h);

	SDL_UnlockSurface(screen);
	SDL_UpdateRect(screen, 0, 0, 0, 0);

	/* Draw borders, put them in list. */
	for (i = 0; i < ARRAY_SIZE(border); i++) {
		border[i].ignore = false;
		border[i].expires.tv_sec = LONG_MAX;
		border[i].expires.tv_usec = 0;
		list_add(&lines, &border[i].list);
		thick_line(screen, &border[i], 0);
	}

	ballsur = ball_surface(screen);

	for (;;) {
		struct coord new, isect, move = ball.move;
		struct line_segment *best_l, *l, *next;
		double best_d;
		SDL_Event event;
		struct timeval now;

		gettimeofday(&now, NULL);

		/* Expire lines */
		list_for_each_safe(&lines, l, next, list) {
			if (timercmp(&l->expires, &now, <)) {
				list_del(&l->list);
				/* FIXME: Undraw properly. */
				thick_line(screen, l, 0xFFFFFFFF);
				free(l);
			}
		}

	again:
		best_d = MAX_X + MAX_Y;
		best_l = NULL;
		new.x = ball.pos.x + move.x;
		new.y = ball.pos.y + move.y;

		list_for_each(&lines, l, list) {
			if (!l->ignore && intersect(&ball.pos, &new, l, &isect)) {
				double d;
				d = dist(&ball.pos, &isect);
				if (d < best_d) {
					best_l = l;
					best_d = d;
				}
			}
		}

		if (best_l) {
			printf("Ball bouncing off (%f,%f %f,%f)!\n",
			       best_l->start.x, best_l->start.y,
			       best_l->end.x, best_l->end.y);
			bounce(&ball, best_l, &move);

			/* If we moved, stop ignoring lines. */
			if (move.x > 0.001 || move.y > 0.001)
				clear_ignore(&lines);

			/* Don't hit the same line twice. */
			printf("Ignoring that line\n");
			best_l->ignore = true;
			goto again;
		}

		if (move.x > 0.001 || move.y > 0.001) {
			clear_ignore(&lines);
			printf("Moving by %f,%f to %f,%f\n", move.x, move.y, new.x, new.y);
		}

		/* Restore what was there before ball. */
		if (savesur) {
			SDL_BlitSurface(savesur, NULL, screen, &rect);
			SDL_UpdateRect(screen, rect.x, rect.y, rect.w, rect.h);
		} else {
			savesur = sub_surface(screen, ballsur->w, ballsur->h);
		}

		/* Save away image under new ball. */
		rect.w = ballsur->w;
		rect.h = ballsur->h;
		rect.x = new.x - ballsur->w/2;
		rect.y = new.y - ballsur->h/2;
		SDL_BlitSurface(screen, &rect, savesur, NULL);

		/* Draw new ball */
		SDL_BlitSurface(ballsur, NULL, screen, &rect);
		SDL_UpdateRect(screen, rect.x, rect.y, rect.w, rect.h);
		/* That SDL_BlitSurface can crop rect. */
		rect.x = new.x - ballsur->w/2;
		rect.y = new.y - ballsur->h/2;
		ball.pos = new;
		SDL_Delay(25);

		while (SDL_PollEvent(&event)) {
			if (event.type == SDL_QUIT) {
				SDL_Quit();
				return 0;
			}
			if (event.type == SDL_MOUSEBUTTONDOWN)
				creating_lines = true;
			else if (event.type == SDL_MOUSEBUTTONUP)
				creating_lines = false;

			if (creating_lines && event.type == SDL_MOUSEMOTION) {
				struct line_segment *n = malloc(sizeof(*n));
				n->start.x = event.motion.x - event.motion.xrel;
				n->start.y = event.motion.y - event.motion.yrel;
				n->end.x = event.motion.x;
				n->end.y = event.motion.y;
				n->ignore = false;
				timeradd(&now, &line_life, &n->expires);
				list_add_tail(&lines, &n->list);
				thick_line(screen, n, 0);
			}
		}
	}
}
#endif
